/*
 *  decoder.cpp
 *  Simple MP3 decoder and ReplayGain analyzer
 *
 *  Created by Alexander Thomas on 21/07/06.
 *  Based on the 'minimad.c' sample code from libMAD.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "gain_analysis.h"
#include "decoder.h"


/* Length of chunks  */
#define BUF_SIZE 16384
#define MYABS(x) (x < 0 ? -x : x)


static FILE *theFile;
static float thePeak;
static unsigned long decoded, decodedLast, maxDecode, numErrors;
static long initFreq = 0;

/*
 * This is a private message structure. A generic pointer to this structure
 * is passed to each of the callback functions. Put here any data you need
 * to access from within the callbacks.
 */
struct buffer {
  unsigned char *start;
  unsigned long length;
};


/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded.
 */
static
enum mad_flow input(void *data,
                    struct mad_stream *stream)
{
  struct buffer *buffer = data;
	unsigned long restLength, newLength, toRead;
	
	/* Move remaining part of buffer to the front */
	restLength = stream->bufend - stream->next_frame;
	/* TODO: find something better than memmove. Moving around memory is a guarantee for slowliness! */
	if( restLength > 0 )
		memmove( buffer->start, buffer->start+buffer->length-restLength, restLength );
	
	/* Fill rest of buffer with new data */
	toRead = BUF_SIZE-restLength;
	if(maxDecode) {
		if( decoded >= maxDecode )
			return MAD_FLOW_STOP;
		else if( decoded+toRead > maxDecode )
			toRead = maxDecode-decoded;
	}
	
	newLength = fread( buffer->start+restLength, 1, toRead, theFile );
	decodedLast = decoded;
	decoded += newLength;
	if( newLength <= 0 )
    return MAD_FLOW_STOP;
	buffer->length = restLength+newLength;
	
#ifdef DEEP_DEBUG
	fprintf( stderr, "Read %lu bytes, buffer: %lu\n", newLength, buffer->length );
#endif
	mad_stream_buffer(stream, buffer->start, buffer->length);
	
  return MAD_FLOW_CONTINUE;
}


/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 */
static inline
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));
	
  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;
	
  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}


/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 */
static
enum mad_flow output(void *data,
                     struct mad_header const *header,
                     struct mad_pcm *pcm)
{
  unsigned int nchannels, nsamples, i = 0;
  mad_fixed_t const *left_ch, *right_ch;
	Float_t *l_samples;
	Float_t *r_samples;
	float newVal;
	
  nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];
	if( pcm->samplerate != initFreq ) {
		if(initFreq)
			fprintf( stderr, "WARNING: sample rate is different from previous file. Album gain will be reset.\n" );
		InitGainAnalysis( pcm->samplerate );
		initFreq = pcm->samplerate;
	}
	
	l_samples = malloc( nsamples*sizeof(Float_t) );
	r_samples = malloc( nsamples*sizeof(Float_t) );
	
#ifdef DEEP_DEBUG
	fprintf( stderr, "Analyzing %u samples\n", nsamples );
#endif
	
	/* GainAnalysis expects samples in the range -32767 ~ 32767 */
  for( i=0; i<nsamples; i++ ) {
		l_samples[i] = (float)(*left_ch++) / (float)(1L << (MAD_F_FRACBITS - 15));
		if( (newVal = MYABS(l_samples[i]) ) > thePeak )
			thePeak = newVal;
		
    if(nchannels == 2) {
			r_samples[i] = (float)(*right_ch++) / (float)(1L << (MAD_F_FRACBITS - 15));;
			if( (newVal = MYABS(r_samples[i]) ) > thePeak )
				thePeak = newVal;
    }
  }
	
	/* TODO: check if it's more efficient to do this on larger chunks of data */
	if( AnalyzeSamples( l_samples, r_samples, nsamples, nchannels ) == GAIN_ANALYSIS_ERROR )
		fprintf( stderr, "Error in AnalyzeSamples\n" ); /* TODO: do something more useful */
	
	free( l_samples );
	free( r_samples );
	
  return MAD_FLOW_CONTINUE;
}


/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static
enum mad_flow error(void *data,
                    struct mad_stream *stream,
                    struct mad_frame *frame)
{
  struct buffer *buffer = data;
	/* TODO: make the offset value absolute, the position within the chunk is useless! */
  fprintf(stderr, "decoding error 0x%04x (%s) at byte offset 0x%lx\n",
	        stream->error, mad_stream_errorstr(stream),
	        decodedLast + (unsigned long)(stream->this_frame-buffer->start));
	numErrors++;
	
  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */
  return MAD_FLOW_CONTINUE;
}


/* A decoder object is instantiated and configured with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error). */
unsigned long decode( FILE *file, unsigned long fSize, float *peak, float *gain )
{
  struct buffer buffer;
  struct mad_decoder decoder;
  int result;
	
	theFile = file;
	maxDecode = fSize;
	
  /* initialize our private message structure */
	
  buffer.start  = malloc(BUF_SIZE);
  buffer.length = 0;
	
  /* configure input, output, and error functions */
	
  mad_decoder_init(&decoder, &buffer,
	                 input, 0 /* header */, 0 /* filter */, output,
	                 error, 0 /* message */);
	
  /* start decoding */
	
	decoded = 0;
	decodedLast = 0;
	thePeak = 0.0;
	numErrors = 0;
	/* We can't do InitGainAnalysis yet because we don't know the sample rate.
	 * The problem is that we need to re-do InitGainAnalysis when the sample rate changes,
	 * resetting the album gain. But since almost all songs on which you'll want to
	 * calculate replaygain will have 44100, this situation shouldn't happen often. */
	
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);
	
  /* release the decoder */
	
  mad_decoder_finish(&decoder);
	free( buffer.start );
	
	*peak = thePeak/32768.0;
	*gain = GetTitleGain();
	
  return numErrors;
}


float getAlbumGain()
{
	return GetAlbumGain();
}
