/*
 * id3gain
 * Copyright (C) 2006-2011 Alexander Thomas (doctor.lex@gmail.com).
 *
 * Version 0.9: 2006/12
 * Version 1.0: 2009/02
 * Version 2.0: 2011/10: direct writing of iTunes-compatible iTunNORM, v1 detection bugfix,
 *              changed -P option, added -A, -J, implemented -f, error checking
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <vector>
using namespace std;
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h> // for utimes()

extern "C" {
#include "decoder.h"
#include "tags.h"
}
#include "main.h"

#define RG_PROGNAME "id3Gain"
#define RG_VERSION "2.0"

//#define DEBUG
//#define DEEPDEBUG


static bool m_bVerbose = false;
static bool m_bDontWrite = false;

void printUsage(const char *pName)
{
	cout << "Usage:\n"
	     << "  " << pName << " [options] input1.mp3 [input2.mp3 ...]\n"
	     << "Options:" << endl;
	cout << " -h: show usage info and help, and exit.\n"
	     << " -a: album mode (calculated album gain is not added if there's only one song)\n"
	     << " -g G: set album gain to G (implies -a)\n"
	     << " -f: re-use, don't recalculate ReplayGain values if already present in the file\n"
	     << " -F: force writing to files even if they cannot be reliably decoded\n"
	     << " -c: remove any replaygain tags (don't calculate gain)\n"
	     << " -C: enable id3v2.4 compression (not compatible with iTunes)\n"
	     << " -N: nuke all ID3 tags (don't calculate gain)\n"
	     << " -p N: if the new tag is larger than the original, use N bytes of padding\n"
	     << "       (default: 128) for the new tag.\n"
	     << " -P N: add padding if necessary to ensure the entire tag is at least N bytes\n"
	     << "       long (overrides -p).\n"
	     << " -1: preserve ID3v1 tag instead of replacing it with new v2 tag\n"
	     << " -2: always erase ID3v1 tag (undoes any preceding -1)\n"
	     << " -I: write or overwrite iTunNORM tag in iTunes-friendly format such that it\n"
	     << "     uses the track (or album if -A) gain instead of Sound Check.\n"
	     << " -i: ignore all other editing options except -A, and update any existing iTunes\n"
	     << "     Sound Check values to track gain (or album gain if -A).\n"
	     << "     You must first run id3gain normally on a file, and then have iTunes read\n"
	     << "     it at least once, to make this work. You may want to use -I instead.\n"     
	     << " -A: use album gain if available for Sound Check (-I or -i).\n"
	     << " -J: remove iTunNORM tag if it is present (overrides -I, -i).\n"
	     << " -d: display only (no editing whatsoever)\n"
	     << " -V: enable verbose mode\n"
	     << " -v: show version and exit" << endl;
}


void process_arguments(int argc, char **argv, vector<char*> &vsFiles,
                       Settings &settings)
{
	// Defaults
	settings.bClean     = false;
	settings.bFast      = false;
	settings.bForce     = false;
	settings.bCompress  = false;
	settings.bReplaceV1 = true;
	settings.bDropV1    = false;
	settings.bNukem     = false;
	settings.nAlbumMode = NOALBUM;
	settings.bSoundCheck    = false;
	settings.bSoundCheckTag = false;
	settings.bSCheckAlbum   = false;
	settings.bKillSCheck    = false;
	settings.fAGain      = 0.0f;
	settings.wXPadding   = 128;
	settings.wMinTagSize = 0;
	
	int i = 0;
	while(++i < argc) {
		if(argv[i][0] != '-')
			vsFiles.push_back(argv[i]);
		else {
			char *arg = argv[i];
			int len = strlen(arg);
			for(int j=1; j<len; j++) {
				switch(arg[j]) {
				case 'h':
					printUsage(argv[0]);
					exit(0);
					break;
				case 'a':
					if(settings.nAlbumMode != ALBUM_USER)
						settings.nAlbumMode = ALBUM_AUTO;
					break;
				case 'A':
					settings.bSCheckAlbum = true;
					break;
				case 'c':
					settings.bClean = true;
					break;
				case 'C':
					settings.bCompress = true;
					break;
				case 'd':
					m_bDontWrite = true;
					break;
				case 'f':
					settings.bFast = true;
					break;
				case 'F':
					settings.bForce = true;
					break;
				case 'g':
					if(i+1 == argc || sscanf(argv[++i], "%g", &settings.fAGain) != 1)
						error_exit("Error: -g expects a floating-point number als next argument.",
						           1, argv[0]);
					settings.nAlbumMode = ALBUM_USER;
					break;
				case 'i':
					settings.bSoundCheck = true;
					break;
				case 'I':
					settings.bSoundCheckTag = true;
					break;
				case 'J':
					settings.bKillSCheck = true;
					break;
				case 'N':
					settings.bNukem = true;
					break;
				case 'p':
					if(i+1 == argc || sscanf(argv[++i], "%lu", &settings.wXPadding) != 1)
						error_exit("Error: -p expects an unsigned integer als next argument.",
						           1, argv[0]);
					break;
				case 'P':
					if(i+1 == argc || sscanf(argv[++i], "%lu", &settings.wMinTagSize) != 1)
						error_exit("Error: -P expects an unsigned integer als next argument.",
						           1, argv[0]);
					break;					
				case '1':
					settings.bReplaceV1 = false;
					break;
				case '2':
					settings.bDropV1    = true;
					settings.bReplaceV1 = true;
					break;
				case 'v':
					cout << RG_PROGNAME << " v" << RG_VERSION << endl;
					exit(0);
					break;
				case 'V':
					m_bVerbose = true;
					break;
				default:
					cerr << "Warning: unknown option -" << arg[j] << endl;
					break;
				}
			}
		}
	}
	
	if(settings.nAlbumMode == ALBUM_AUTO && vsFiles.size() < 2) {
		settings.nAlbumMode = NOALBUM;
		if(m_bVerbose)
			cout << "Only one input file: ignoring -a switch." << endl;
	}
	
	if(settings.bKillSCheck) {
		settings.bSoundCheck = false;
		settings.bSoundCheckTag = false;
	}
}


int main(int argc, char *argv[])
{
	Settings settings;
	unsigned long bufLength, fileLength;
	float aPeak=0.0;
	FILE *theFile;
	vector<char*> vsFiles;
	
	/* parse arguments */
	process_arguments(argc, argv, vsFiles, settings);
	unsigned int numFiles = vsFiles.size();
	if(numFiles == 0)
		error_exit("Not enough arguments", 1, argv[0]);

	if(settings.bSoundCheck) {
		fixSoundCheck(vsFiles, settings.bSCheckAlbum);
		return 0;
	}

	vector<long>  vnID3Lengths(numFiles, 0);
	vector<float> vfPeakValues(numFiles), vfGainValues(numFiles);
	vector<bool>  vbHasV1(numFiles, false), vbHasV2(numFiles, false);
	
	
	/* Step 1: analyze all files */
	for(unsigned int f=0; f<numFiles; f++) {
		if(m_bVerbose)
			cout << "File: " << vsFiles[f] << endl;
		bool bHasV1=false, bHasV2=false; // impossible to pass a vector<bool> item as reference
		getID3Info(vsFiles[f], vnID3Lengths[f], bHasV1, bHasV2);
		vbHasV1[f] = bHasV1;
		vbHasV2[f] = bHasV2;
		
		if(settings.bClean || settings.bNukem)
			continue;
		
		if(settings.bFast && bHasV2) {
			// See if the tags are already there
			unsigned char *tagBuff = new unsigned char[vnID3Lengths[f]];
			bufLength = readFirstPart(tagBuff, vnID3Lengths[f], vsFiles[f]);
			if(bufLength < (unsigned long)vnID3Lengths[f])
				cerr << "Warning: premature end of file, or tag is b0rked" << endl;			
			if(tagBuff) {
				bool bHaveValues = false;
				double dGain, dPeak, dAGain, dAPeak;
				struct id3_tag *fileTag = id3_tag_parse(tagBuff, bufLength);
				
				if(fileTag == NULL)
					cerr << "ERROR: could not parse id3 tag!" << endl;
				else if(getReplayTags(fileTag, dGain, dPeak, dAGain, dAPeak)) {
					/* Conditions for not having to analyze the file:
					 * 1. Both gain & peak values are present
					 * 2. If album gain is requested and there is more than 1 file,
					 *    both album gain & peak are present */
					if(dPeak > 0.0 &&
					   (numFiles == 1 || settings.nAlbumMode != ALBUM_AUTO || dAPeak > 0.0)) {
						vfGainValues[f] = dGain;
						vfPeakValues[f] = dPeak;
						bHaveValues = true;
						cout << "Re-using existing ReplayGain values" << endl;
					}
				}
				id3_tag_delete(fileTag);
				if(bHaveValues) {
					continue;
				}
			}
		}
		
		fileLength = getFileSize(vsFiles[f]);
		unsigned long decodeLength = fileLength;
		theFile = fopen(vsFiles[f], "rb");
		if(theFile == NULL)
			error_exit("Could not open input file", 1, NULL);
		
		if(vnID3Lengths[f] > 0) {
			fseek(theFile, vnID3Lengths[f], SEEK_SET);
			decodeLength -= vnID3Lengths[f];
		}
		if(vbHasV1[f])
			decodeLength -= 128;
		
		/*==== This is where the actual analysis happens... ====*/
		cout << "Analyzing " << vsFiles[f] << "..." << endl;
		unsigned long wNumErrors = decode(theFile, decodeLength,
		                                  &(vfPeakValues[f]), &(vfGainValues[f]));
		fclose(theFile);
		
		/* Sanity check (I once mp3gained mp3gain itself...) */
		if(!settings.bForce && wNumErrors > 24) {
			cerr << "ERROR: Skipping this file because it has " << wNumErrors << " decoding errors.\n"
			     << "       It may not be an MP3 file. Use -F to write anyway." << endl;
			vfPeakValues[f] = -1.0f;
			continue;
		}
		if(m_bVerbose && wNumErrors > 0)
			cerr << "Decoding errors: " << wNumErrors << endl;
		
		printf("  Recommended dB change for song: %+.2f dB\n", vfGainValues[f]);
		printf("  Peak value: %.6f\n", vfPeakValues[f]);
		if(settings.nAlbumMode != NOALBUM) {
			float recAGain = getAlbumGain();
			if(settings.nAlbumMode == ALBUM_AUTO)
				settings.fAGain = recAGain;
			if(vfPeakValues[f] > aPeak)
				aPeak = vfPeakValues[f];
			printf("  Recommended album gain up to this file: %+.2f dB\n", recAGain);
			printf("  Album peak up to this file: %.6f\n", aPeak);
		}
	}
		
		
	/* Step 2: write (or nuke) the ID3 tags */
	if(settings.bNukem) {
		for(unsigned int f=0; f<numFiles; f++) {
			if(!(vbHasV1[f] || vbHasV2[f]))
				continue;
			if(m_bVerbose) {
				cout << "Stripping ";
				if(vbHasV1[f] && vbHasV2[f]) cout << "v1 and v2";
				else if(vbHasV1[f]) cout << "v1";
				else cout << "v2";
				cout << " from file " << vsFiles[f] << endl;
			}
			if(m_bDontWrite)
				continue;
			
			fileLength = getFileSize(vsFiles[f]);
			unsigned long xLength = 0;
			int fd = open(vsFiles[f], O_RDWR);
			if(vbHasV2[f]) {
				char *addr = (char*)mmap(0, fileLength, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
				memmove(addr, addr+vnID3Lengths[f], fileLength-vnID3Lengths[f]);
				munmap(addr, fileLength);
				xLength = vnID3Lengths[f];
			}
			if(vbHasV1[f])
				xLength += 128;
			
			if(ftruncate(fd, fileLength-xLength))
				cerr << "WARNING: Error while truncating file!" << endl;
			close(fd);
		}
		return 0;
	}
	
	for(unsigned int f=0; f<numFiles; f++) {
		struct id3_tag *fileTag = NULL;
		unsigned char  *tagBuff = NULL;

		if(vfPeakValues[f] < 0)
			continue;
		
		fileLength = getFileSize(vsFiles[f]);
		
		if(vbHasV2[f]) {
			tagBuff = new unsigned char[vnID3Lengths[f]];
			bufLength = readFirstPart(tagBuff, vnID3Lengths[f], vsFiles[f]);
			if(bufLength < (unsigned long)vnID3Lengths[f])
				cerr << "Warning: premature end of file, or tag is b0rked" << endl;
		}
		else if(vbHasV1[f]) {
			tagBuff = new unsigned char[128];
			bufLength = readLastPart(tagBuff, 128, vsFiles[f]);
		}
		
		if(tagBuff) {
			fileTag = id3_tag_parse(tagBuff, bufLength);
			// TODO: I noticed that this can fail if the iTunNORM field is the last in the tag.
			// The parser may get confused due to the unexpected 0x00 byte at the end.
			// A workaround is to add an extra field in iTunes and delete it afterwards.
			// A solution could be to convert this byte to a space and re-try parsing the tag.
			if(fileTag == NULL)
				cerr << "ERROR: could not parse id3 tag!" << endl;
#ifdef DEBUG
			else {
				tagInfo(fileTag);
#ifdef DEEPDEBUG
				tagFramesInfo(fileTag);
#endif //DEEPDEBUG
			}
#endif //DEBUG
		}
		else {
			/* TODO: Include option to leave files with no tags found untouched? */
			if(m_bVerbose) cout << "File has no ID3 tag" << endl;
			fileTag = id3_tag_new();
		}			
		
		if(settings.bClean) {
			int r = 0;
			r += deleteTXXX(fileTag, "replaygain_track_gain");
			r += deleteTXXX(fileTag, "replaygain_track_peak");
			r += deleteTXXX(fileTag, "replaygain_album_gain");
			r += deleteTXXX(fileTag, "replaygain_album_peak");
			if(r == 0) { /* Nothing has changed */
				if(m_bVerbose) cout << "No existing replaygain fields found." << endl;
				delete [] tagBuff;
				id3_tag_delete(fileTag);
				continue;
			}
		}
		else {
			/*==== Updating the ID3 tag ====*/
			char stringBuf[256];
			sprintf(stringBuf, "%+.2f dB", vfGainValues[f]);
			setTXXX(fileTag, "replaygain_track_gain", stringBuf);
			sprintf(stringBuf, "%.6f", vfPeakValues[f]);
			setTXXX(fileTag, "replaygain_track_peak", stringBuf);
			if(settings.nAlbumMode != NOALBUM) {
				sprintf(stringBuf, "%+.2f dB", settings.fAGain);
				setTXXX(fileTag, "replaygain_album_gain", stringBuf);
				sprintf(stringBuf, "%.6f", aPeak);
				setTXXX(fileTag, "replaygain_album_peak", stringBuf);
			}
			/* Could also include 'RGAD' tag, as proposed by replayGain.
			 * But, the bit format for that tag is so ill-defined that it's probably useless. */
			/*id3_tag_attachframe(fileTag, makeID3BinaryFrame("RGAD", "brolbrol", 8));*/
			
			if(settings.bKillSCheck) {
				if(deleteCOMM(fileTag, "iTunNORM") && m_bVerbose)
					cout << "Stripped iTunNORM from " << vsFiles[f] << endl;
			}
			else if(settings.bSoundCheckTag) {
				const char *lang  = "eng";
				const char *label = "iTunNORM";
				char normTag[96]; // Strictly spoken the string should be 90 bytes
				unsigned int nGain1000, nGain2500, nPeak;
				// These are 'plausible' bogus values for whatever values #5,6,9,10 are supposed to be
				unsigned int nMysteryA1 = 0x20000, nMysteryA2 = 0x20000,
				             nMysteryB1 = 0x3000,  nMysteryB2 = 0x3000;
				
				string sExisting = getITunNORM(fileTag);
				if(sExisting != "") {
					// Yes, I'm anal enough to preserve those values if they are present
					sscanf(sExisting.substr(37,8).c_str(), "%08X", &nMysteryA1);
					sscanf(sExisting.substr(46,8).c_str(), "%08X", &nMysteryA2);
					sscanf(sExisting.substr(73,8).c_str(), "%08X", &nMysteryB1);
					sscanf(sExisting.substr(82,8).c_str(), "%08X", &nMysteryB2);
					if(nMysteryA2 == 0) {
						nMysteryA2 = nMysteryA1;
						nMysteryB2 = nMysteryB1;
					}
				}
				
				iTunizeGains((settings.bSCheckAlbum ? settings.fAGain : vfGainValues[f]),
							 (settings.bSCheckAlbum ? aPeak : vfPeakValues[f]),
				             nGain1000, nGain2500, nPeak);
				// Mind the space at the end. It is not supposed to be there, see below...
				snprintf(normTag, 96,
						 " %08X %08X %08X %08X %08X %08X %08X %08X %08X %08X ",
						 nGain1000, nGain1000,
						 nGain2500, nGain2500,
						 nMysteryA1, nMysteryA2, 
						 nPeak, nPeak,
						 nMysteryB1, nMysteryB2);
				if(setCOMM(fileTag, lang, label, normTag)) {
					if(m_bVerbose)
						cout << "Replaced existing iTunNORM tag" << endl;
				}
				else if(m_bVerbose)
					cout << "Added new iTunNORM tag" << endl;
			}
		}
				
		/* Leaving this option on can produce painful results. Please people, let the id3v1 tag die. */
		id3_tag_options(fileTag, ID3_TAG_OPTION_ID3V1, 0);
		if(!settings.bCompress)
			id3_tag_options(fileTag, ID3_TAG_OPTION_COMPRESSION, 0);
		/* NOTE: when converting pre-v4 tags with genre fields, the genre may show up as a number
		 * in iTunes. This is iTunes' fault, because even in v4 tags it expects genre numbers
		 * to be between brackets, while these were omitted from the standard since v4.
		 * This is why iTunes users should convert all tags to v4 before using this program,
		 * and back to v3 or lower to avoid ending up with non-standard junk. */
		
		if(settings.bSoundCheckTag) {
			// Make the tag more iTunes-like. This does not include the incorrect
			// extra 0x00 bytes after the fields of course...
			id3_tag_options(fileTag, ID3_TAG_OPTION_COMPRESSION, 0);
			id3_tag_options(fileTag, ID3_TAG_OPTION_CRC, 0);
		}
		
		/*==== Rendering the ID3 tag ====*/
		unsigned long id3_newLength = id3_tag_render(fileTag, NULL);
#ifdef DEBUG
		cerr << "New id3 tag length: " << id3_newLength << endl;
#endif
		
		if(settings.wMinTagSize > 0) {
			if(id3_newLength < settings.wMinTagSize) {
				id3_newLength = settings.wMinTagSize;
				fileTag->paddedsize = id3_newLength;
			}
		}
		else if(id3_newLength > (unsigned long)vnID3Lengths[f] && settings.wXPadding > 0) {
			id3_newLength += settings.wXPadding;
			fileTag->paddedsize = id3_newLength;
#ifdef DEBUG
			cerr << "Added padding for total length of: " << id3_newLength << endl;
#endif
		}
		
		delete [] tagBuff;
		tagBuff = new unsigned char[id3_newLength];
		if(id3_tag_render(fileTag, tagBuff) != id3_newLength) {
#ifdef DEBUG
			cerr << "UGH! Final tag is not same size as predicted!" << endl;
#endif
		}
	
		if(m_bDontWrite) {
			delete [] tagBuff;
			id3_tag_delete(fileTag);
			continue;
		}
		
		/* And now, let's pull a rabbit from our hat!
		 * Due to iTunes' flawed ID3 implementation, it expects a NULL byte after
		 * the iTunNORM field. If this byte is not there, iTunes ignores the field and
		 * will either rewrite the entire tag or pretend there is no tag at all.
		 * That's why we're now looking up that space we inserted a while ago and
		 * convert it to 0x00. */
		if(settings.bSoundCheckTag) {
			char pFindMe[16] = {0,0,0,'e','n','g','i','T','u','n','N','O','R','M',0,' '};
			// 160 = 3+[engiTunNORM]+1+[ hex stuff]+1
			for(unsigned long i=0; i<id3_newLength-106; i++) {
				if(memcmp(pFindMe, &(tagBuff[i]), 16) == 0) {
					if(tagBuff[i+105] == ' ')
						tagBuff[i+105] = 0x00;
					else
						cerr << "Unexpected error while iTunesifying iTunNORM!" << endl;
					break;
				}
			}
		}
		
		/*==== Writing the ID3 tag (back) to the file ====*/
		if(id3_newLength == (unsigned long)vnID3Lengths[f]) {
			/* If tag length is unchanged, just open with mmap and update tag. */
#ifdef DEBUG
			cerr << "New tag is same size as old, overwriting using mmap" << endl;
#endif
			int fd = open(vsFiles[f], O_RDWR);
			char *addr = (char*)mmap(0, fileLength, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
			memcpy(addr, (char*)tagBuff, id3_newLength);
			
			munmap(addr, fileLength);
			close(fd);
			/* Touching file to update modification date
			 * (apparently mmap editing has no effect on it). */
			utimes(vsFiles[f], NULL);
		}
		else if(id3_newLength < (unsigned long)vnID3Lengths[f]) {
			/* If tag is N bytes shorter, open with mmap, write new tag, and shift data N bytes to front.
			 * Use ftruncate to cut off the N last bytes of the file. */
			unsigned int xLength = vnID3Lengths[f]-id3_newLength;
#ifdef DEBUG
			cerr << "New tag is shorter than old (-" << xLength << "), shifting data forward using mmap and truncating" << endl;
#endif
			int fd = open(vsFiles[f], O_RDWR);
			char *addr = (char*)mmap(0, fileLength, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
			memcpy(addr, (char*)tagBuff, id3_newLength);
			memmove(addr+id3_newLength, addr+vnID3Lengths[f], fileLength-vnID3Lengths[f]);
			
			munmap(addr, fileLength);
			if(ftruncate(fd, fileLength-xLength))
				cerr << "WARNING: Error while truncating file!" << endl;
			close(fd);
		}
		else {
			/* If tag is M bytes longer, open in r/w-append mode, and append M last bytes after file. Close,
			 * re-open with mmap, and shift data M bytes to back. Write new tag in front. */
			unsigned int xLength = id3_newLength-vnID3Lengths[f];
#ifdef DEBUG
			cerr << "New tag is longer than old (+" << xLength << "), shifting data backward using mmap and appending" << endl;
#endif
			char *xBuff = new char[xLength];
			int fd = open(vsFiles[f], O_RDWR | O_APPEND);
			char *addr = (char*)mmap(0, fileLength, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
			memcpy(xBuff, addr+fileLength-xLength, xLength);
			memmove(addr+xLength, addr, fileLength-xLength);
			write(fd, xBuff, xLength);
			memcpy(addr, (char*)tagBuff, id3_newLength);
			
			munmap(addr, fileLength);
			close(fd);
			delete [] xBuff;
		}
		
		/*==== Erase the V1 tag if required ====*/
		if(vbHasV1[f] && ((settings.bReplaceV1 && !vbHasV2[f]) || settings.bDropV1)) {
			if(m_bVerbose) cout << "Erasing old id3v1 tag." << endl;
			fileLength = getFileSize(vsFiles[f]);
			if(truncate(vsFiles[f], fileLength-128))
				cerr << "WARNING: Error while truncating file!" << endl;
		}
		
		id3_tag_delete(fileTag);
		delete [] tagBuff;
	}

	return 0;
}


void error_exit(const char *msg, int val, const char *pName)
{
	cerr << msg << endl;
	if(pName)
		printUsage(pName);
	exit(val);
}


void getID3Info(const char* filePath,
                long &id3_length, bool &bHasV1, bool &bHasV2)
{
	unsigned char firstBuf[2048], lastBuf[128];
		
	/* Using id3_file_open and id3_file_tag appears to be useless, because it
	 * gives wrong information about the tag version and size. Or, I just have
	 * no idea how to use it correctly due to the total absence of documentation.
	 * So I do it the low-level way, which gives a lot more control too. */
	unsigned long bufLength = readFirstPart(firstBuf, 2048, filePath);
	if(bufLength == 0)
		error_exit("File could not be read or is empty.", 1, NULL);
		
	readLastPart(lastBuf, 128, filePath);
	if(id3_tag_query(lastBuf, 128) == 128) {
		if(m_bVerbose) cout << "File has an ID3v1 tag." << endl;
		bHasV1 = true;
	}
	/* TODO: allow ID3v2 footers: use negative numbers */
	id3_length = id3_tag_query(firstBuf, bufLength);
	if(id3_length > 0) {
		if(m_bVerbose) cout << "File has an ID3v2 tag." << endl;
		bHasV2 = true;
	}
}


void iTunizeGains(double dGain, double dPeak,
                  unsigned int &nGain1000, unsigned int &nGain2500, unsigned int &nPeak)
{
	nGain1000 = (unsigned int)round(pow(10, -dGain/10.0)*1000);
	nGain2500 = (unsigned int)round(pow(10, -dGain/10.0)*2500);
	nPeak =     (unsigned int)round(dPeak*32768);
	if(nGain1000 > 65535)
		nGain1000 = 65535;
	if(nGain2500 > 65535)
		nGain2500 = 65535;
	if(nPeak > 32768)
		nPeak = 32768;
}


bool getReplayTags(struct id3_tag *fileTag,
                   double &dGain, double &dPeak, double &dAGain, double &dAPeak)
{	
	bool bResult = false;
	struct id3_frame *pFrame;
	unsigned char *tmp;
	dPeak = dAPeak = -1.0;
	
	pFrame = findTXXX(fileTag, "replaygain_track_gain");
	if(pFrame != NULL) {
		bResult = true;
		tmp = getField(id3_frame_field(pFrame, 2));
		sscanf((char*)tmp, "%lf ", &dGain);
		free(tmp);
		dPeak = 0.0;
	}
	
	pFrame = findTXXX(fileTag, "replaygain_album_gain");
	if(pFrame != NULL) {
		bResult = true;
		tmp = getField(id3_frame_field(pFrame, 2));
		sscanf((char*)tmp, "%lf ", &dAGain);
		free(tmp);
		dAPeak = 0.0;
	}
	if(!bResult)
		return false;

	pFrame = findTXXX(fileTag, "replaygain_track_peak");
	if(pFrame) {
		tmp = getField(id3_frame_field(pFrame, 2));
		sscanf((char*)tmp, "%lf", &dPeak);
		free(tmp);
	}

	pFrame = findTXXX(fileTag, "replaygain_album_peak");
	if(pFrame) {
		tmp = getField(id3_frame_field(pFrame, 2));
		sscanf((char*)tmp, "%lf", &dAPeak);
		free(tmp);
	}
	
	return bResult;
}


string getITunNORM(struct id3_tag *fileTag)
{
	struct id3_frame *scFrame = findCOMM(fileTag, "iTunNORM");
	if(scFrame == NULL)
		return string();
	
	unsigned char *tmp = getField(id3_frame_field(scFrame, 3));
	string sSCValues((char*)tmp);
	free(tmp);

	return sSCValues;
}


void fixSoundCheck(const vector<char*> &vsFiles, bool bAlbum)
{
	unsigned int numFiles = vsFiles.size();
	for(unsigned int f=0; f<numFiles; f++) {
		if(m_bVerbose)
			cout << "File: " << vsFiles[f] << endl;
		bool bHasV1, bHasV2;
		long id3Length;
		getID3Info(vsFiles[f], id3Length, bHasV1, bHasV2);	
		if(!bHasV2) {
			cerr << "Skipping `" << vsFiles[f] << "': has no v2 tag" << endl;
			continue;
		}

		// Read tag
		struct id3_tag *fileTag = NULL;
		unsigned char  *tagBuff = new unsigned char[id3Length];
		
		unsigned long bufLength = readFirstPart(tagBuff, id3Length, vsFiles[f]);
		if(bufLength < (unsigned long)id3Length)
			cerr << "Warning: premature end of file, or tag is b0rked" << endl;
		
		fileTag = id3_tag_parse(tagBuff, bufLength);
		if(fileTag == NULL) {
			cerr << "Skipping `" << vsFiles[f] << "': could not parse id3 tag" << endl;
			delete [] tagBuff;
			continue;
		}

		// Get RG values
		double dGain, dPeak, dAGain, dAPeak;
		if(!getReplayTags(fileTag, dGain, dPeak, dAGain, dAPeak)) {
			cerr << "Skipping `" << vsFiles[f] << "': no ReplayGain tags found" << endl;
			delete [] tagBuff;
			id3_tag_delete(fileTag);
			continue;
		}
		if(bAlbum || dPeak < 0.0) {
			// if bAlbum, prefer album values, fallback to track
			// if !bAlbum, fallback to album values if no track values
			if(dAPeak >= 0.0) {
				dGain = dAGain;
				dPeak = dAPeak;
			}
		}
#ifdef DEEPDEBUG
		cout << "Using gain: " << dGain << " and peak " << dPeak << endl;
#endif

		// convert to SC values
		unsigned int nGain1000, nGain2500, nPeak;
		iTunizeGains(dGain, dPeak, nGain1000, nGain2500, nPeak);
		
		char pGain1000[9], pGain2500[9], pPeak[9];
		snprintf(pGain1000, 9, "%08X", nGain1000);
		snprintf(pGain2500, 9, "%08X", nGain2500);
		snprintf(pPeak, 9, "%08X", nPeak);
		string sSCValues = getITunNORM(fileTag);
		delete [] tagBuff;
		id3_tag_delete(fileTag);

		if(sSCValues == "") {
			cerr << "Skipping `" << vsFiles[f] << "': no existing iTunNORM tag found" << endl;
			continue;
		}
#ifdef DEEPDEBUG
		cout << "Existing iTunNORM values: `" << sSCValues << "'" << endl;
#endif
		
		string sOldSCValues(sSCValues);
		sSCValues.replace(1,8,pGain1000); // beware of the stupid leading space
		sSCValues.replace(19,8,pGain2500);
		sSCValues.replace(55,8,pPeak);
		if(sSCValues.substr(10,8) != "00000000") {
			sSCValues.replace(10,8,pGain1000);
			sSCValues.replace(28,8,pGain2500);
			sSCValues.replace(64,8,pPeak);
		}
#ifdef DEEPDEBUG
		cout << "New values: `" << sSCValues << "'" << endl;
#endif

		if(sSCValues == sOldSCValues) {
			if(m_bVerbose)
				cout << "iTunNORM tag is already up-to-date" << endl;
			continue;
 		}
		if(m_bDontWrite) {
			cout << "Would update Sound Check to gain of " << dGain << "dB and peak of " << nPeak << endl;
			continue;
		}

		// Open file in mmap mode (only need to map the tag area)
		int fd = open(vsFiles[f], O_RDWR);
		char *addr = (char*)mmap(0, bufLength, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);

		// Find and update iTunNORM hex codes
		char findMe[102];
		strcpy(findMe,"engiTunNORM");
		memcpy(findMe+12, sOldSCValues.c_str(), 90);
		long tagPos = memfind(addr, bufLength, findMe, 102);
		if(tagPos < 0) {
			cerr << "Error: could not locate iTunNORM field for writing, this is not supposed to happen" << endl;
			continue;
		}
		else
			memcpy(addr+12+tagPos, sSCValues.c_str(), 90);
		munmap(addr, bufLength);
		close(fd);
		if(tagPos >= 0) 
			utimes(vsFiles[f], NULL); // 'touch' the file
		if(m_bVerbose)
			cout << "Updated Sound Check to gain of " << dGain << "dB and peak of " << nPeak << endl;
	}
}

/*--- File routines ---*/

unsigned long getFileSize(const char *fName)
{
	struct stat buffer;
	if(stat(fName, &buffer) != 0) {
		cerr << "Error while opening file for updating!" << endl;
		exit(1);
	}
	return buffer.st_size;
}		


int readFirstPart(unsigned char *buff, unsigned int length,
                  const char *fileName)
{
	int len;
	FILE *file = fopen(fileName, "rb");
	if(file == NULL)
		return -1;
	len = fread(buff, 1, length, file);
	fclose(file);
	
	return len;
}

int readLastPart(unsigned char *buff, int length,
                 const char *fileName)
{
	int len;
	FILE *file = fopen(fileName, "rb");
	if(file == NULL)
		return -1;
	if(fseek(file, -length, SEEK_END)) {
		fclose(file);
		return -1;
	}
	len = fread(buff, 1, length, file);
	fclose(file);
	
	return len;
}


long memfind(const char *mem,   unsigned long memSize,
             const char *query, unsigned long querySize)
{
	if(memSize < querySize)
		return -1;
	memSize -= querySize-1;
	for(unsigned long i=0; i<=memSize; i++) {
		bool ok=true;
		for(unsigned long j=0; j<querySize; j++) {
			if(mem[i+j] != query[j]) {
				ok = false;
				break;
			}
		}
		if(ok)
			return i;
	}
	return -1;
}
